/**
  This code for the arduino was written by KE4EST, Michael Graves, starting in May of 2016.
  The purpose of this code is to decode the DiSEqC commands issued by most FTA receivers.
  This is Version 0.4.6 and is constantly being updated.
  Feel free to use this code for your own project. All I ask is that you leave a reference to myself, KE4EST.
  This code has been tested on the Arduino, Nano, Uno, and the Mega.
  A couple of line need to be changed depending on which board is being used.

**/
#include <LiquidCrystal.h>
#define rawBtn A5

#define detectReg 7 // for Mega
//#define detectReg 5 // for Uno, Nano
LiquidCrystal lcd(12, 11, 7, 6, 5, 4);

volatile int pwm_value = 0; //used inside interrupt for microseconds between pulses
volatile int prev_time = 0; //used with pwm_value to grab the start time

volatile bool nibble1[4]; //array to hold each nibble
volatile bool nextLineOK = 1;  //OK to start counter for next line
volatile int bitCount = 0;  // used to count the bits(when you hit 4 you have a nibble that can be processed)
volatile int ninthBitCount = 0; //Used to look for 9th bit to ignore

volatile bool packetDetect = 0; //Used to let toneCount know it can start counting
volatile int toneCount = 0; //Used to count to flash LED to indicate tone has been detected
int byteParse = 0; //used to group the output into two's so that, E0316B3C look like E0 31 6B 3C
int nextPacket = 0; //increments with each full nibble, to know when 8 is achevied so it knows end of packet has been reached

unsigned long purge = 0;  // used to count after packet, to issue a CR

String diseqc = ""; //String to store the commands as string, used to parse data for displaying what the commands are
//                    Swtich port 1, or Motor goto 16, etc.
String rawInput; //String to store the raw input ie. 1110 0000 0011 0001......
String hexOutput;  //String to test for repeats
long test = -1;
bool insertZero = 0;
bool errorCheck = 1;
byte newChar[8] = { B00000, B00000, B01010, B11111, B11111, B11111, B01110, B00100 };
unsigned long errorCount = 0;


void setup(void)
{
  //  digitalWrite(11, 1);
  lcd.begin(20, 4);
  lcd.createChar(0, newChar);
  pinMode(rawBtn, INPUT);
  pinMode(13, OUTPUT);
  Serial.begin(115200);
  lcd.setCursor(7, 0);
  lcd.print("Arduino");
  lcd.setCursor(3, 1);
  lcd.print("DISEQC Decoder");
  lcd.setCursor(0, 2);
  lcd.print("     By: KE4EST");
  lcd.setCursor(0, 3);
  lcd.print("Ver. 0.4.6  06.22.16");
  Serial.println("KE4EST Arduino DiSEqC Decoder");
  Serial.println("Version 0.4.6");
  Serial.println("June 22nd 2016");
  delay(2000);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("DISEQC: ");
  lcd.setCursor(0, 1);
  lcd.print("Cmd Type:  ");
  lcd.setCursor(0, 2);
  lcd.print("Data");
  //  lcd.setCursor(0, 3);
  //  lcd.print("");
  //Serial.println();

  attachInterrupt(0, rising, RISING); // when pin D2 goes high, call the rising function


}

void loop(void)
{

  if (packetDetect)
  {
    PORTB |= (1 << detectReg); // Tone detect LED attached to pin 13

    lcd.setCursor(19, 3);
    lcd.write((byte)0);
    toneCount++;
    if (toneCount == 256)
    {
      PORTB &= ~(1 << detectReg);
      lcd.setCursor(19, 3);
      lcd.print(" ");
      toneCount = 0;
      packetDetect = 0;
    }
  }

  purge++;

  if (ninthBitCount == 9)
  {
    //rawInput -= " ";
    bitCount = 0;
    ninthBitCount = 0;
  }
  if (bitCount == 4)
  {
    //rawInput += " ";

    outputNibble(nibble1);
    if (insertZero)
    {
      bitCount = 1;
      nibble1[0] = 0;
      ninthBitCount++;
      insertZero = 0;
    }
    else
    {
      bitCount = 0;
    }
    byteParse += 1;
    if (byteParse == 2)
    {
      diseqc += " ";
      byteParse = 0;
    }
    nextPacket = nextPacket + 1;
  }
  if (nextPacket == 8)
  {
    if (diseqc != "")
    {
      if (digitalRead(rawBtn) == 1)
      {
        Serial.println();
        Serial.print("Raw: ");
        Serial.println(rawInput);
        test += 1;
      }
      else
      {
        Serial.println();
        test += 1;
        Serial.print(test);
        Serial.print("::");
      }


      Serial.print("Hex:: ");
      Serial.print(diseqc);
      lcd.setCursor(8, 0);
      lcd.print(diseqc);
      if (hexOutput == diseqc)
      {
        Serial.println(" ::Repeated Command");
        lcd.setCursor(0, 3);
        lcd.print("Repeated ");
        lcd.print(test);
        if (test == 1)
        {
          lcd.print(" Time");
        }
        else
        {
          lcd.print(" Times");
        }
      }
      else
      {
        test = 0;
        hexOutput = diseqc;
        Serial.println();
        lcd.setCursor(0, 3);
        lcd.print("                   ");
      }
    }
    if (digitalRead(rawBtn) == 1)
    {
      String raw1 = rawInput.substring(0, 19);
      String raw2 = rawInput.substring(20, 39);
      lcd.setCursor(0, 1);
      lcd.print(raw1);
      lcd.setCursor(0, 2);
      lcd.print(raw2);
    }
    else
    {
      DATA_EXTRACT(diseqc);
    }
    rawInput = "";
    diseqc = "";
    nextPacket = 0;
    errorCheck = 1;
    EIFR = 1;
    //bitCount = 0;

  }

  if (purge > 1024 && diseqc != "")
  {
    Serial.print(test);
    Serial.println();
    Serial.print("Hex:: ");
    Serial.print(diseqc);
    lcd.setCursor(17, 0);
    lcd.print("  ");
    lcd.setCursor(8, 0);
    lcd.print(diseqc);
    if (hexOutput == diseqc)
    {
      Serial.println(" ::Repeated Command");
      lcd.setCursor(0, 3);
      lcd.print("Repeated ");
      lcd.print(test);
      if (test == 1)
      {
        lcd.print(" Time");
      }
      else
      {
        lcd.print(" Times");
      }
    }
    else
    {
      test = 0;
      hexOutput = diseqc;
      Serial.println();
      lcd.setCursor(0, 3);
      lcd.print("                   ");
    }
    if (digitalRead(rawBtn) == 1)
    {
      String raw1 = rawInput.substring(0, 19);
      String raw2 = rawInput.substring(20, 39);
      lcd.setCursor(0, 1);
      lcd.print(raw1);
      lcd.setCursor(0, 2);
      lcd.print(raw2);
    }
    else
    {
      DATA_EXTRACT(diseqc);
    }
    rawInput = "";
    diseqc = "";
    nextPacket = 0;
    purge = 0;
    bitCount = 0;
    EIFR = 1;

  }
  errorCheck = 1;
}

void rising(void)
{
  attachInterrupt(0, falling, FALLING);
  prev_time = micros();

}

void falling(void)
{
  attachInterrupt(0, rising, RISING);
  packetDetect = 1;
  pwm_value = micros() - prev_time;
  //  Serial.print(pwm_value); //uncomment for testing
  //  Serial.print(" ");
  if (pwm_value > 300 && pwm_value < 600)
  {
    nibble1[bitCount] = 0;
    bitCount++;
    ninthBitCount++;
  }
  if (pwm_value > 600 && pwm_value < 900)
  {
    nibble1[bitCount] = 1;
    bitCount++;
    ninthBitCount++;
  }
}

void outputNibble(volatile bool nibble[4])
{
  purge = 0;


  if (nibble[0] == 0 && nibble[1] == 0 && nibble[2] == 0 && nibble[3] == 0)
  {
    diseqc += "0";
    rawInput += "0000 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 0 && nibble[2] == 0 && nibble[3] == 1)
  {
    diseqc += "1";
    rawInput += "0001 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 0 && nibble[2] == 1 && nibble[3] == 0)
  {
    diseqc += "2";
    rawInput += "0010 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 0 && nibble[2] == 1 && nibble[3] == 1)
  {
    //Serial.print("3");
    diseqc += "3";
    rawInput += "0011 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 1 && nibble[2] == 0 && nibble[3] == 0)
  {
    diseqc += "4";
    rawInput += "0100 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 1 && nibble[2] == 0 && nibble[3] == 1)
  {
    diseqc += "5";
    rawInput += "0101 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 1 && nibble[2] == 1 && nibble[3] == 0)
  {
    diseqc += "6";
    rawInput += "0110 ";
  }
  else if (nibble[0] == 0 && nibble[1] == 1 && nibble[2] == 1 && nibble[3] == 1)
  {
    diseqc += "7";
    rawInput += "0111 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 0 && nibble[2] == 0 && nibble[3] == 0)
  {
    diseqc += "8";
    rawInput += "1000 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 0 && nibble[2] == 0 && nibble[3] == 1)
  {
    //Serial.print("9");
    diseqc += "9";
    rawInput += "1001 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 0 && nibble[2] == 1 && nibble[3] == 0)
  {
    diseqc += "A";
    rawInput += "1010 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 0 && nibble[2] == 1 && nibble[3] == 1)
  {
    //Serial.print("B");
    diseqc += "B";
    rawInput += "1011 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 1 && nibble[2] == 0 && nibble[3] == 0)
  {
    if (errorCheck)
    {
      Serial.println("Error Caught!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
      errorCount ++;
      lcd.setCursor(16, 2);
      lcd.print(errorCount);
      diseqc += "E";
      rawInput += "1110 ";
      errorCheck = 0;
      insertZero = 1;
    }
    else
    {
      diseqc += "C";
      rawInput += "1100 ";
    }
  }
  else if (nibble[0] == 1 && nibble[1] == 1 && nibble[2] == 0 && nibble[3] == 1)
  {
    diseqc += "D";
    rawInput += "1101 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 1 && nibble[2] == 1 && nibble[3] == 0)
  {
    diseqc += "E";
    rawInput += "1110 ";
  }
  else if (nibble[0] == 1 && nibble[1] == 1 && nibble[2] == 1 && nibble[3] == 1)
  {
    diseqc += "F";
    rawInput += "1111 ";
  }
}


void DATA_EXTRACT(String data)
{
  String diseqcParse = data.substring(3, 5);
  //  Serial.print("Sub String Test = ");
  //  Serial.println(diseqcParse);
  if (diseqcParse == "31")
  {
    Serial.print("Motor Command: ");
    lcd.setCursor(0, 1);
    lcd.print("Cmd Type: Motor    ");
    String diseqcParse = data.substring(6, 8);
    String mNumber = data.substring(9, 11);
    Motor(diseqcParse, mNumber);
  }
  else if (diseqcParse == "10")
  {
    Serial.print("Switch Command: ");
    lcd.setCursor(0, 1);
    lcd.print("Cmd Type: Switch  ");
    String diseqcParse = data.substring(6, 8);
    String swNumber = data.substring(9, 11);
    Switch(diseqcParse, swNumber);
  }
  else
  {
    Serial.println("Unrecognized Command!!");
    lcd.setCursor(0, 1);
    lcd.print("Decode Error!! ");
  }
}


void Motor(String mData, String mNum)
{
  if (mData == "6B")
  {
    Serial.print("Goto: ");
    Serial.println(hexStoDec(mNum));
    lcd.setCursor(0, 2);
    lcd.print("                    ");
    lcd.setCursor(0, 2);
    lcd.print("Goto: ");
    lcd.print(hexStoDec(mNum));
  }
  if (mData == "6A")
  {
    Serial.print("Store As Position: ");
    Serial.println(hexStoDec(mNum));
    lcd.setCursor(0, 2);
    lcd.print("Store As Pos: ");
    lcd.print(hexStoDec(mNum));
  }
  if (mData == "69")
  {
    Serial.print("Move West: ");
    Serial.println(commandLookUp(mNum));
    lcd.setCursor(0, 2);
    lcd.print("Go West ");
    lcd.print(commandLookUp(mNum));
  }
  if (mData == "68")
  {
    Serial.print("Move East: ");
    Serial.println(commandLookUp(mNum));
    lcd.setCursor(0, 2);
    lcd.print("Go East ");
    lcd.print(commandLookUp(mNum));
  }
  if (mData == "60")
  {
    Serial.println("Motor Stop!!");
    lcd.setCursor(0, 2);
    lcd.print("STOP!!!            ");
  }
  if (mData == "63")
  {
    Serial.println("Disable Limits");
  }
  if (mData == "66")
  {
    Serial.println("Set East Limit");
  }
  if (mData == "67")
  {
    Serial.println("Set West Limit");
  }
  if (mData == "6B" && mNum == "00")
  {
    Serial.println("Goto Reference");
  }
  if (mData == "67")
  {
    Serial.println("Set West Limit");
  }
  if (mData == "6F")
  {
    Serial.println("(Re)Calculate Posit");
  }
}


void Switch(String sData, String sNum)
{
  if (sData == "38")
  {
    Serial.print("DiSEqC 1.0, Port ");
    Serial.println(swCommCommandLookUp(sNum));
  }
  if (sData == "39")
  {
    Serial.print("DiSEqC 1.1, Port ");
    Serial.println(swUncommCommandLookUp(sNum));
  }
}


unsigned int hexStoDec(String hexString)
{
  unsigned int decValue = 0;
  int nextInt;

  for (int i = 0; i < hexString.length(); i++) {

    nextInt = int(hexString.charAt(i));
    if (nextInt >= 48 && nextInt <= 57) nextInt = map(nextInt, 48, 57, 0, 9);
    if (nextInt >= 65 && nextInt <= 70) nextInt = map(nextInt, 65, 70, 10, 15);
    if (nextInt >= 97 && nextInt <= 102) nextInt = map(nextInt, 97, 102, 10, 15);
    nextInt = constrain(nextInt, 0, 15);

    decValue = (decValue * 16) + nextInt;
  }

  return decValue;
}


String commandLookUp(String command)
{
  if (command == "FE")
  {
    return "Nudge";
  }
  if (command == "00")
  {
    return "Continuous";
  }
}


int swUncommCommandLookUp(String command)
{
  if (command == "F0")
  {
    return 1;
  }
  if (command == "F1")
  {
    return 2;
  }
  if (command == "F2")
  {
    return 3;
  }
  if (command == "F3")
  {
    return 4;
  }
  if (command == "F4")
  {
    return 5;
  }
  if (command == "F5")
  {
    return 6;
  }
  if (command == "F6")
  {
    return 7;
  }
  if (command == "F7")
  {
    return 8;
  }
  if (command == "F8")
  {
    return 9;
  }
  if (command == "F9")
  {
    return 10;
  }
  if (command == "FA")
  {
    return 11;
  }
  if (command == "FB")
  {
    return 12;
  }
  if (command == "FC")
  {
    return 13;
  }
  if (command == "FD")
  {
    return 14;
  }
  if (command == "FE")
  {
    return 15;
  }
  if (command == "FF")
  {
    return 16;
  }
}


int swCommCommandLookUp(String command)
{
  if (command == "F0" || command == "F2")
  {
    return 1;
  }
  if (command == "F4" || command == "F6")
  {
    return 2;
  }
  if (command == "F8" || command == "FA")
  {
    return 3;
  }
  if (command == "FC" || command == "FE")
  {
    return 4;
  }
}

